#include <node.h>
#include <iostream>
#include <vexplib.h>


namespace vexp {

using v8::Function;
using v8::FunctionCallbackInfo;
using v8::Isolate;
using v8::Exception;
using v8::Local;
using v8::Object;
using v8::String;
using v8::Value;
using v8::Number;
using v8::Integer;
using v8::Boolean;
using v8::Array;
using v8::Persistent;
using v8::Context;


// the native vexp library
VEXPLib accessor;



// ----------------------------------------------------------------------------
// sqw index tree functions (@author tweber)
// ----------------------------------------------------------------------------
/**
 * Allocate an sqw index tree
 */
void createSqwTree(const FunctionCallbackInfo<Value>& args) {
	auto *iso = args.GetIsolate();

	if(args.Length() != 0) {
		auto str = String::NewFromUtf8(iso, "createSqwTree takes no arguments");
		iso->ThrowException(Exception::TypeError(str.ToLocalChecked()));
		return;
	}

	int handle = accessor.createSqwTree();
	args.GetReturnValue().Set(Integer::New(iso, handle));
}


/**
 * Deallocate an sqw index tree
 */
void removeSqwTree(const FunctionCallbackInfo<Value>& args) {
	auto *iso = args.GetIsolate();

	if(args.Length() != 1 || !args[0]->IsInt32()) {
		auto str = String::NewFromUtf8(iso, "removeSqwTree takes one argument: the tree handle");
		iso->ThrowException(Exception::TypeError(str.ToLocalChecked()));
		return;
	}

	int handle = Local<Integer>::Cast(args[0])->Value();
	accessor.removeSqwTree(handle);
}


/**
 * Clear an sqw index tree
 */
void clearSqwTree(const FunctionCallbackInfo<Value>& args) {
	auto *iso = args.GetIsolate();

	if(args.Length() != 1 || !args[0]->IsInt32()) {
		auto str = String::NewFromUtf8(iso, "clearSqwTree takes one argument: the tree handle");
		iso->ThrowException(Exception::TypeError(str.ToLocalChecked()));
		return;
	}

	int handle = Local<Integer>::Cast(args[0])->Value();
	accessor.clearSqwTree(handle);
}


/**
 * Add a node to an sqw index tree
 */
void addSqwValue(const FunctionCallbackInfo<Value>& args) {
	auto *iso = args.GetIsolate();

	if(args.Length() != 6 || !args[0]->IsInt32() || 
		!args[1]->IsNumber() || !args[2]->IsNumber() || !args[3]->IsNumber() || !args[4]->IsNumber() || !args[5]->IsNumber()) {
		auto str = String::NewFromUtf8(iso, "addSqwValue takes six arguments: the tree handle and h, k, l, E, S");
		iso->ThrowException(Exception::TypeError(str.ToLocalChecked()));
		return;
	}

	int handle = Local<Integer>::Cast(args[0])->Value();

	double h = Local<Number>::Cast(args[1])->Value();
	double k = Local<Number>::Cast(args[2])->Value();
	double l = Local<Number>::Cast(args[3])->Value();
	double E = Local<Number>::Cast(args[4])->Value();
	double S = Local<Number>::Cast(args[5])->Value();

	accessor.addSqwValue(handle, h, k, l, E, S);
}


/**
 * Load an sqw index tree from an ascii table
 */
void loadSqwTable(const FunctionCallbackInfo<Value>& args) {
	auto *iso = args.GetIsolate();

	if(args.Length() != 2 || !args[0]->IsInt32() || !args[1]->IsString()) {
		auto str = String::NewFromUtf8(iso, "loadSqwTable takes two arguments: the tree handle and the file name");
		iso->ThrowException(Exception::TypeError(str.ToLocalChecked()));
		return;
	}

	int handle = Local<Integer>::Cast(args[0])->Value();

	Local<String> str = Local<String>::Cast(args[1]);

	int len = str->Utf8Length(iso);
	if(len < 0) {
		std::cerr << "Invalid string argument in loadSqwTable!" << std::endl;
		args.GetReturnValue().Set(Boolean::New(iso, false));
		return;
	}

	char *file = new char[len+1];
	file[len] = 0;
	str->WriteUtf8(iso, file, len+1);

	//const char *file = *String::Utf8Value(iso, args[1]);

	bool ok = accessor.loadSqwTable(handle, file);
	args.GetReturnValue().Set(Boolean::New(iso, ok));

	delete[] file;
}



/**
 * Query for the nearest node in an sqw index tree
 */
void querySqwTree(const FunctionCallbackInfo<Value>& args) {
	auto *iso = args.GetIsolate();
	Local<Context> context = iso->GetCurrentContext();

	if(args.Length() != 5 || !args[0]->IsInt32() || 
		!args[1]->IsNumber() || !args[2]->IsNumber() || !args[3]->IsNumber() || !args[4]->IsNumber()) {
		auto str = String::NewFromUtf8(iso, "querySqwTree takes five arguments: the tree handle and the h, k, l, E coordinates");
		iso->ThrowException(Exception::TypeError(str.ToLocalChecked()));
		return;
	}

	int handle = Local<Integer>::Cast(args[0])->Value();
	double h = Local<Number>::Cast(args[1])->Value();
	double k = Local<Number>::Cast(args[2])->Value();
	double l = Local<Number>::Cast(args[3])->Value();
	double E = Local<Number>::Cast(args[4])->Value();
	double S = 0;

	bool bOk = accessor.querySqwTree(handle, h, k, l, E, S);

	auto result = Object::New(iso);
	result->Set(context, String::NewFromUtf8(iso, "ok").ToLocalChecked(), Boolean::New(iso, bOk));
	result->Set(context, String::NewFromUtf8(iso, "h").ToLocalChecked(), Number::New(iso, h));
	result->Set(context, String::NewFromUtf8(iso, "k").ToLocalChecked(), Number::New(iso, k));
	result->Set(context, String::NewFromUtf8(iso, "l").ToLocalChecked(), Number::New(iso, l));
	result->Set(context, String::NewFromUtf8(iso, "E").ToLocalChecked(), Number::New(iso, E));
	result->Set(context, String::NewFromUtf8(iso, "S").ToLocalChecked(), Number::New(iso, S));

	args.GetReturnValue().Set(result);
}
// ----------------------------------------------------------------------------




// ----------------------------------------------------------------------------
// CIF handling functions (@author tweber)
// ----------------------------------------------------------------------------
/**
 * Allocate a CIF object
 */
void createCIF(const FunctionCallbackInfo<Value>& args) {
	auto *iso = args.GetIsolate();

	if(args.Length() != 0) {
		auto str = String::NewFromUtf8(iso, "createCIF takes no arguments");
		iso->ThrowException(Exception::TypeError(str.ToLocalChecked()));
		return;
	}

	int handle = accessor.createCIF();
	args.GetReturnValue().Set(Integer::New(iso, handle));
}


/**
 * Deallocate a CIF object
 */
void removeCIF(const FunctionCallbackInfo<Value>& args) {
	auto *iso = args.GetIsolate();

	if(args.Length() != 1 || !args[0]->IsInt32()) {
		auto str = String::NewFromUtf8(iso, "removeCIF takes one argument: the CIF handle");
		iso->ThrowException(Exception::TypeError(str.ToLocalChecked()));
		return;
	}

	int handle = Local<Integer>::Cast(args[0])->Value();
	accessor.removeCIF(handle);
}


/**
 * Clear a CIF object
 */
void clearCIF(const FunctionCallbackInfo<Value>& args) {
	auto *iso = args.GetIsolate();

	if(args.Length() != 1 || !args[0]->IsInt32()) {
		auto str = String::NewFromUtf8(iso, "clearCIF takes one argument: the CIF handle");
		iso->ThrowException(Exception::TypeError(str.ToLocalChecked()));
		return;
	}

	int handle = Local<Integer>::Cast(args[0])->Value();
	accessor.clearCIF(handle);
}


/**
 * Load a CIF
 */
void loadCIF(const FunctionCallbackInfo<Value>& args) {
	auto *iso = args.GetIsolate();

	if(args.Length() != 3 || !args[0]->IsInt32() || !args[1]->IsString() || !args[2]->IsBoolean()) {
		auto str = String::NewFromUtf8(iso, "loadCIF takes three arguments: the CIF handle, the file name or data string, and a boolean");
		iso->ThrowException(Exception::TypeError(str.ToLocalChecked()));
		return;
	}

	int handle = Local<Integer>::Cast(args[0])->Value();

	Local<String> str = Local<String>::Cast(args[1]);

	int len = str->Utf8Length(iso);
	if(len < 0) {
		std::cerr << "Invalid string argument in loadCIF!" << std::endl;
		args.GetReturnValue().Set(Boolean::New(iso, false));
		return;
	}

	char *file = new char[len+1];
	file[len] = 0;
	str->WriteUtf8(iso, file, len+1);

	//const char* file = *String::Utf8Value(iso, args[1]->ToString());
	bool is_file = Local<Boolean>::Cast(args[2])->Value();

	//const char* file = "/Users/tw/tmp/mnsi.cif";
	//bool is_file = true;

	bool ok = accessor.loadCIF(handle, file, is_file);
	args.GetReturnValue().Set(Boolean::New(iso, ok));

	delete[] file;
}


/**
 * Get the lattice parameters from the CIF
 */
void getCIFLattice(const FunctionCallbackInfo<Value>& args) {
	auto *iso = args.GetIsolate();
	Local<Context> context = iso->GetCurrentContext();

	if(args.Length() != 1 || !args[0]->IsInt32()) {
		auto str = String::NewFromUtf8(iso, "getCIFLattice takes one argument: the CIF handle");
		iso->ThrowException(Exception::TypeError(str.ToLocalChecked()));
		return;
	}

	int handle = Local<Integer>::Cast(args[0])->Value();

	double a=0, b=0, c=0;
	double alpha=0, beta=0, gamma=0;
	bool bOk = accessor.getCIFLattice(handle, a, b, c, alpha, beta, gamma);

	auto result = Object::New(iso);
	result->Set(context, String::NewFromUtf8(iso, "ok").ToLocalChecked(), Boolean::New(iso, bOk));
	result->Set(context, String::NewFromUtf8(iso, "a").ToLocalChecked(), Number::New(iso, a));
	result->Set(context, String::NewFromUtf8(iso, "b").ToLocalChecked(), Number::New(iso, b));
	result->Set(context, String::NewFromUtf8(iso, "c").ToLocalChecked(), Number::New(iso, c));
	result->Set(context, String::NewFromUtf8(iso, "alpha").ToLocalChecked(), Number::New(iso, alpha));
	result->Set(context, String::NewFromUtf8(iso, "beta").ToLocalChecked(), Number::New(iso, beta));
	result->Set(context, String::NewFromUtf8(iso, "gamma").ToLocalChecked(), Number::New(iso, gamma));

	args.GetReturnValue().Set(result);
}


/**
 * Get the number of atoms from the CIF
 */
void getCIFNumAtoms(const FunctionCallbackInfo<Value>& args) {
	auto *iso = args.GetIsolate();
	
	if(args.Length() != 1 || !args[0]->IsInt32()) {
		auto str = String::NewFromUtf8(iso, "getCIFNumAtoms takes one argument: the CIF handle");
		iso->ThrowException(Exception::TypeError(str.ToLocalChecked()));
		return;
	}

	int handle = Local<Integer>::Cast(args[0])->Value();
	int num = accessor.getCIFNumAtoms(handle);
	args.GetReturnValue().Set(Integer::New(iso, num));
}


/**
 * Get an atom position from the CIF
 */
void getCIFAtom(const FunctionCallbackInfo<Value>& args) {
	auto *iso = args.GetIsolate();
	Local<Context> context = iso->GetCurrentContext();

	if(args.Length() != 2 || !args[0]->IsInt32() || !args[1]->IsInt32()) {
		auto str = String::NewFromUtf8(iso, "getCIFAtom takes two arguments: the CIF handle and the atom index");
		iso->ThrowException(Exception::TypeError(str.ToLocalChecked()));
		return;
	}

	int handle = Local<Integer>::Cast(args[0])->Value();
	int atomnr = Local<Integer>::Cast(args[1])->Value();

	double x=0, y=0, z=0;
	const char* atomname = accessor.getCIFAtom(handle, atomnr, x, y, z);

	auto result = Object::New(iso);
	result->Set(context, String::NewFromUtf8(iso, "ok").ToLocalChecked(), Boolean::New(iso, atomname!=0));
	result->Set(context, String::NewFromUtf8(iso, "name").ToLocalChecked(), String::NewFromUtf8(iso, atomname).ToLocalChecked());
	result->Set(context, String::NewFromUtf8(iso, "x").ToLocalChecked(), Number::New(iso, x));
	result->Set(context, String::NewFromUtf8(iso, "y").ToLocalChecked(), Number::New(iso, y));
	result->Set(context, String::NewFromUtf8(iso, "z").ToLocalChecked(), Number::New(iso, z));

	args.GetReturnValue().Set(result);
}


/**
 * Get the number of generated atoms from the CIF
 */
void getCIFNumGeneratedAtoms(const FunctionCallbackInfo<Value>& args) {
	auto *iso = args.GetIsolate();

	if(args.Length() != 2 || !args[0]->IsInt32() || !args[1]->IsInt32()) {
		auto str = String::NewFromUtf8(iso, "getCIFNumGeneratedAtoms takes two arguments: the CIF handle and the atom index");
		iso->ThrowException(Exception::TypeError(str.ToLocalChecked()));
		return;
	}

	int handle = Local<Integer>::Cast(args[0])->Value();
	int atomnr = Local<Integer>::Cast(args[1])->Value();
	int num = accessor.getCIFNumGeneratedAtoms(handle, atomnr);

	args.GetReturnValue().Set(Integer::New(iso, num));
}


/**
 * Get a generated atom position from the CIF
 */
void getCIFGeneratedAtom(const FunctionCallbackInfo<Value>& args) {
	auto *iso = args.GetIsolate();
	Local<Context> context = iso->GetCurrentContext();

	if(args.Length() != 3 || !args[0]->IsInt32() || !args[1]->IsInt32() || !args[2]->IsInt32()) {
		auto str = String::NewFromUtf8(iso, "getCIFGeneratedAtom takes three arguments: the CIF handle, the base atom index and the generated atom index");
		iso->ThrowException(Exception::TypeError(str.ToLocalChecked()));
		return;
	}

	int handle = Local<Integer>::Cast(args[0])->Value();
	int atomnr = Local<Integer>::Cast(args[1])->Value();
	int gennr = Local<Integer>::Cast(args[2])->Value();

	double x=0, y=0, z=0;
	const char* atomname = accessor.getCIFGeneratedAtom(handle, atomnr, gennr, x, y, z);

	auto result = Object::New(iso);
	result->Set(context, String::NewFromUtf8(iso, "ok").ToLocalChecked(), Boolean::New(iso, atomname!=0));
	result->Set(context, String::NewFromUtf8(iso, "name").ToLocalChecked(), String::NewFromUtf8(iso, atomname).ToLocalChecked());
	result->Set(context, String::NewFromUtf8(iso, "x").ToLocalChecked(), Number::New(iso, x));
	result->Set(context, String::NewFromUtf8(iso, "y").ToLocalChecked(), Number::New(iso, y));
	result->Set(context, String::NewFromUtf8(iso, "z").ToLocalChecked(), Number::New(iso, z));

	args.GetReturnValue().Set(result);
}


/**
 * Get the number of symmetry operations from the CIF
 */
void getCIFNumSymOps(const FunctionCallbackInfo<Value>& args) {
	auto *iso = args.GetIsolate();

	if(args.Length() != 1 || !args[0]->IsInt32()) {
		auto str = String::NewFromUtf8(iso, "getCIFNumSymOps takes one argument: the CIF handle");
		iso->ThrowException(Exception::TypeError(str.ToLocalChecked()));
		return;
	}

	int handle = Local<Integer>::Cast(args[0])->Value();
	int num = accessor.getCIFNumSymOps(handle);
	args.GetReturnValue().Set(Integer::New(iso, num));
}


/**
 * Get a symmetry operation from the CIF
 */
void getCIFSymOp(const FunctionCallbackInfo<Value>& args) {
	auto *iso = args.GetIsolate();
	Local<Context> context = iso->GetCurrentContext();

	if(args.Length() != 2 || !args[0]->IsInt32() || !args[1]->IsInt32()) {
		auto str = String::NewFromUtf8(iso, "getCIFSymOp takes two arguments: the CIF handle and the symop index");
		iso->ThrowException(Exception::TypeError(str.ToLocalChecked()));
		return;
	}

	int handle = Local<Integer>::Cast(args[0])->Value();
	int atomnr = Local<Integer>::Cast(args[1])->Value();

	double m00=1, m01=0, m02=0, m03=0;
	double m10=0, m11=1, m12=0, m13=0;
	double m20=0, m21=0, m22=1, m23=0;
	double m30=0, m31=0, m32=0, m33=1;
	bool ok = accessor.getCIFSymOp(handle, atomnr, 
		m00, m01, m02, m03,
		m10, m11, m12, m13,
		m20, m21, m22, m23,
		m30, m31, m32, m33);

	auto result = Object::New(iso);

	auto op = Array::New(iso, 16);
	op->Set(context, 0, Number::New(iso, m00)); op->Set(context, 1, Number::New(iso, m01)); op->Set(context, 2, Number::New(iso, m02)); op->Set(context, 3, Number::New(iso, m03));
	op->Set(context, 4, Number::New(iso, m10)); op->Set(context, 5, Number::New(iso, m11)); op->Set(context, 6, Number::New(iso, m12)); op->Set(context, 7, Number::New(iso, m13));
	op->Set(context, 8, Number::New(iso, m20)); op->Set(context, 9, Number::New(iso, m21)); op->Set(context, 10, Number::New(iso, m22)); op->Set(context, 11, Number::New(iso, m23));
	op->Set(context, 12, Number::New(iso, m30)); op->Set(context, 13, Number::New(iso, m31)); op->Set(context, 14, Number::New(iso, m32)); op->Set(context, 15, Number::New(iso, m33));

	result->Set(context, String::NewFromUtf8(iso, "ok").ToLocalChecked(), Boolean::New(iso, ok));
	result->Set(context, String::NewFromUtf8(iso, "op").ToLocalChecked(), op);

	args.GetReturnValue().Set(result);
}

// ----------------------------------------------------------------------------




/**
 * The init function declares what we will make visible to node.
 */
void init(Local<Object> exports) {

	// ------------------------------------------------------------------------
	// Register sqw functions
	NODE_SET_METHOD(exports, "createSqwTree", createSqwTree);
	NODE_SET_METHOD(exports, "removeSqwTree", removeSqwTree);
	NODE_SET_METHOD(exports, "clearSqwTree", clearSqwTree);

	NODE_SET_METHOD(exports, "addSqwValue", addSqwValue);
	NODE_SET_METHOD(exports, "loadSqwTable", loadSqwTable);
	NODE_SET_METHOD(exports, "querySqwTree", querySqwTree);
	// ------------------------------------------------------------------------


	// ------------------------------------------------------------------------
	// Register CIF functions
	NODE_SET_METHOD(exports, "createCIF", createCIF);
	NODE_SET_METHOD(exports, "removeCIF", removeCIF);
	NODE_SET_METHOD(exports, "clearCIF", clearCIF);

	NODE_SET_METHOD(exports, "loadCIF", loadCIF);
	NODE_SET_METHOD(exports, "getCIFLattice", getCIFLattice);

	NODE_SET_METHOD(exports, "getCIFNumAtoms", getCIFNumAtoms);
	NODE_SET_METHOD(exports, "getCIFAtom", getCIFAtom);

	NODE_SET_METHOD(exports, "getCIFNumGeneratedAtoms", getCIFNumGeneratedAtoms);
	NODE_SET_METHOD(exports, "getCIFGeneratedAtom", getCIFGeneratedAtom);

	NODE_SET_METHOD(exports, "getCIFNumSymOps", getCIFNumSymOps);
	NODE_SET_METHOD(exports, "getCIFSymOp", getCIFSymOp);
	// ------------------------------------------------------------------------
}


NODE_MODULE(vexplibAddon, init);

}
