const TrajectoryLine = require('./trajectory-line');
const LineSegment = require('./line-segment');
const {scale, vector3WorldToView, scalarWorldToView, vector3ViewToWorld} = require('../conversion');
const Color = require('../colors');
const {mapColor} = require('../map-color');
const { mode } = require('../../server-state');
const { clientSettings } = require('../../client-state');
const THREE = require('three');

class EnergyLine extends TrajectoryLine {
    constructor(type, parent, start, end, center, steps) {
        super(type, parent, start, end, center, steps);

        // Material.
        this._lineMaterial = new THREE.MeshBasicMaterial({ color: Color.Q, depthTest: true, transparent: false });

        // The data stored.
        this._data = [];

        // The line is a segment.
        this._line = null;

        // Create a group for the data.
        this._dataGroup = new THREE.Group();
        this._group.add(this._dataGroup);

        // Group for the data spheres.
        this._dataSphereGroup = new THREE.Group();
        this._dataGroup.add(this._dataSphereGroup);

        // Group for the point spheres.
        this._pointSphereGroup = new THREE.Group();
        this._dataGroup.add(this._pointSphereGroup);

        this.highlight(false);
        this.updateLine();
    }
    
    setData(data) {

        // Update only if the new data has more points.
        if (data.length > this._data.length) {
            this._data = data;
        }
    }

    getEnergyScale() {

        // The scale is only for ki and kf trajectories which are orthogonal to the scattering plane.
        // Scale is applied for the direction orthogonal to the plane.
        if (this._type == 'ki' || this._type == 'kf') {
            return clientSettings.energyScale;
        }
        return 1;
    }

    removeSpheres() {

        // Remove the data spheres.
        for (let i = this._dataSphereGroup.children.length - 1; i >= 0; i--) {
            this._dataSphereGroup.remove(this._dataSphereGroup.children[i]);
        }

        // Remove the point spheres.
        for (let i = this._pointSphereGroup.children.length - 1; i >= 0; i--) {
            this._pointSphereGroup.remove(this._pointSphereGroup.children[i]);
        }
    }

    updateDataGroup() {
        
        // Reset the group transform.
        let inv = new THREE.Matrix4();
        inv.getInverse(this._dataGroup.matrix);
        this._dataGroup.applyMatrix(inv);

        // Calculate the rotation matrix between (start, end) and (0, 0, 1)
        let matrix4 = this.calculateSegmentRotationMatrix(this._start, this._end);

        // Calculate the position of the new frame.
        let start = this._start.clone();

        // Apply the scale in the z direction.
        start.setZ(start.z * clientSettings.energyScale);
        matrix4.setPosition(start);

        // Then we can apply the new matrix.
        this._dataGroup.applyMatrix(matrix4);
        
        // Remove the old spheres.
        this.removeSpheres();
        
        // Calculate length of the segment. Applies the scale.
        let length = this._end.clone().sub(this._start).length() * this.getEnergyScale();

        // Calculate the data sphere radius.
        let dataSphereRadius = clientSettings.dataSphereRadius * 0.5 / scale;

        // Calculate delta between spheres and the z-radius of the spheres.
        // There is a special case for trajectories with 1 point i.e. 0 intervals.
        let delta = 0;
        let zSphereRadius = 0;
        if (this._intervals > 0) {
            delta = length / this._intervals;
            zSphereRadius = 0.5 * delta;
        }
        else {
            zSphereRadius = dataSphereRadius;
        }
        
        // Set the scale for the data spheres.
        let dataSphereScale = dataSphereRadius / zSphereRadius;

        this._dataSphereGroup.scale.x = dataSphereScale;
        this._dataSphereGroup.scale.y = dataSphereScale;

        // Iterate the points and draw data sphere or point sphere.
        for (let i = 0; i < this._intervals + 1; i++) {

            let position = new THREE.Vector3(0, 0, i * delta);
            let intensity = this._data[i];

            // Draw a data sphere.
            // Do not draw the sphere if intensity < 0 i.e. point has been skipped.
            if (i < this._data.length && intensity >= 0) {
                let dataMaterial = new THREE.MeshBasicMaterial({color: 0xffffff, depthTest: true, transparent: false});
                const color = mapColor(intensity, 0, clientSettings.maxIntensity);
                dataMaterial.color.setRGB(color.r / 255, color.g / 255, color.b / 255);
                let sphere = new THREE.Mesh(this.sphereGeometry, dataMaterial);
                sphere.scale.multiplyScalar(zSphereRadius);
                sphere.renderOrder = 4;
                sphere.position.copy(position);
                this._dataSphereGroup.add(sphere);
            }
            // Draw a point sphere.
            else {
                let pointMaterial = new THREE.MeshBasicMaterial({color: 0xffffff, depthTest: true, transparent: false});
                let sphere = new THREE.Mesh(this.sphereGeometry, pointMaterial);
                sphere.scale.multiplyScalar(1 / scale);
                sphere.renderOrder = 4;
                sphere.position.copy(position);
                this._pointSphereGroup.add(sphere);
            }
        }
    }

    /**
     * Update line position and delta spheres.
     */
    updateLine() {

        // The line segment is not displayed in spy mode.
        if (!mode.spyMode) {

            // Remove the line from the data group.
            if (this._line !== null) {
                this._dataGroup.remove(this._line.object3D);
                // Destroy the line segment.
                this._line.destroy();
            }

            // Calculate length of the segment. Applies the scale.
            let length = this._end.clone().sub(this._start).length() * this.getEnergyScale();
            
            // Create a line segment for the line.
            if (length > 0) {
                // It is necessary not to draw line with length 0 because the underlying cylinder has a default length 1.
                this._line = new LineSegment(new THREE.Vector3(0, 0, 0), new THREE.Vector3(0, 0, length), 0.5 / scale, this._lineMaterial);

                // Add the line to the data group.
                this._dataGroup.add(this._line.object3D);
            }
            else {
                this._line = null;
            }
        }

        // Update the data.
        this.updateDataGroup();
    }

    highlightLines(color) {
    }

    highlightSpheres(value) {

        // The spheres are always visible in spy mode.
        if (!mode.spyMode) {
            this._dataSphereGroup.visible = value;
            this._pointSphereGroup.visible = value;
        }
        else {
            this._dataSphereGroup.visible = true;
            this._pointSphereGroup.visible = true;
        }
    }

    /**
     * Set line visibility.
     */
    setVisibility(visibility) {

        if (this._line !== null) {
            this._line.object3D.visible = visibility;
        }
        this._dataGroup.visible = visibility;
        this._pointSphereGroup.visible = visibility;
    }

    /**
     * Remove this trajectory properly.
     */
    destroy() {
        // Remove from the parent.
        this._parent.remove(this._group);

        // Destroy the line segment.
        if (this._line !== null) {
            this._line.destroy();
        }

        // Remove the spheres.
        this.removeSpheres();
    }
}

module.exports = EnergyLine;