const THREE = require('three');
const os = require('os');
const fs = require('fs');
const path = require('path');
const { exec } = require('child_process');
const View3D = require('../view3d');
const { peaks, sample, matrix } = require('../../../server-state');
const {vector3WorldToView} = require('../../../utils/conversion');
const xtal = require('../../../../server/model/xtal/xtal');
const default_config = require("../../../../../resources/default-config");


class BzView3D extends View3D{
    constructor(scene)
    {
        super(scene);

	let tmpdir = os.tmpdir();
        this._fileCodeName = path.join(tmpdir, this.getRandomeFileCodeName());

        this._scene = scene;
        this._mainGroupe = new THREE.Group();
        this._drawTemplate = [];

        this._geometry = new THREE.Geometry();
        this._jsonData;
        
        this._bzViewMode = "single";
        this._showFace = true;
        this._showEdge = false;
        this._normalColorMode = "absolut";
        this._colorTemplate = "unicolor";

        this._faceMaterial;
        this._edgeMaterial;

        this._faceColor = [0, 0, 75, 0.15];
        this._edgeColor = [0, 0, 0, 1];
        this._facesIndex;

        this._offsetPosition = [0, 0, 0];

        this._takinPath = default_config["takin_bz"];
        this.recreateGeometry();
    }

    ////////////////////////////////////////////////////////////
    ////        This part is for all Main function
    ////////////////////////////////////////////////////////////

    /**
     * Create bz zone from scratch
     * it will automaticly call changeBzStruc after
     */
    recreateGeometry()
    {
        this.getNewGeometryData();
    }

    /**
     * Geometry data does not change
     * this function will not call takin
     */
    changeBzStruct()
    {
        this._scene.remove(this._mainGroupe);

        this._mainGroupe = new THREE.Group();
        this._mainGroupe.name = "main_bz";

        this._polygonGroupe = new THREE.Group();

        //create the mesh and material

        if(this._showEdge)
        {
            this.createEdgeMaterial();
            this.createEdgeMesh();
        }
        if(this._showFace)
        {
            this.createFaceMaterial();
            this.createFaceMesh();
        }


        if(this._bzViewMode == "single")
        {
            this.drawOne(this._offsetPosition);
        }
        else if(this._bzViewMode == "multiple")
        {
            this.drawMutiple();
        }

        this._scene.add(this._mainGroupe);
    }

    ////////////////////////////////////////////////////////////
    ////        This part is for all that can be see
    ////////////////////////////////////////////////////////////


    /**
     * This function will create face mesh
     */
    createFaceMesh()
    {
        let face = new THREE.Mesh(this._geometry, this._faceMaterial);
        this._polygonGroupe.add(face);
    }
    
    /**
     * This function will create mesh edge outline
     */
    createEdgeMesh()
    {
        let edge = new THREE.EdgesGeometry(this._geometry);
        let line = new THREE.LineSegments(edge, this._edgeMaterial);
        this._polygonGroupe.add(line);
    }

    /**
     * This function convert rgba to THREE.Color() (alpha are not use)
     * @param {*} rgba a array wit max value (255, 255, 255)
     * @returns THREE.Color
     */
    createColor(rgba)
    {
        return new THREE.Color(rgba[0] / 255, rgba[1]/ 255, rgba[2] / 255);
    }

    /**
     * This function will create edge material
     * return => this._edgeMaterial
     */
    createEdgeMaterial()
    {
        this._edgeMaterial = new THREE.LineBasicMaterial({
            color: this.createColor(this._edgeColor), transparent:true, opacity:this._edgeColor[3] 
        })
    }

    /**
     * This function create face material
     * return => this._faceMaterial
     */
    createFaceMaterial()
    {
        let material;

        if(this._normalColorMode =="relative" && this._colorTemplate == "normal")
        {
            
            material = new THREE.MeshNormalMaterial({
                transparent:true, opacity:this._faceColor[3]
            })
        }
        else if(this._normalColorMode == "absolut" && this._colorTemplate == "normal")
        {
            material = new THREE.MeshBasicMaterial({
                vertexColors:true, transparent:false
            })
            this.colorEachFace();
        }
        else
        {
            material = new THREE.MeshStandardMaterial({
                color: this.createColor(this._faceColor), transparent:true, opacity:this._faceColor[3],
            })
        }
        this._faceMaterial = material;
        
    }

    /**
     * This function create a color for each face
     */
    colorEachFace()
    {
        this._geometry.vertexColors = true;
        let geometryFaces = this._geometry.faces;
        let facesIndex = this._facesIndex;
        let facesIndexNumber = facesIndex.length;
        let counter = 0
        for (let x = 0; x < facesIndexNumber; x++) {
            let faceGroupe = this._facesIndex[x];

            for (let y = 0; y < faceGroupe.length; y++) {
                geometryFaces[counter].vertexColors = true;
                geometryFaces[counter].color.set(new THREE.Color().setHSL(x / facesIndexNumber, 1, 0.75));
                counter ++;
            }
        }
        this._geometry.colorsNeedUpdate = true;

    }

    /**
     * This function draw one instance of polygonGroup to mainGroup
     * return => polygonGroup
     * @param {[x,y,z]} position undefinded position = [0, 0, 0];
     */
    drawOne(position)
    {
        let polygonGroupe = this._polygonGroupe.clone();

        if(position == undefined)
        {
            position = [0, 0, 0];
        }

        let vector3 = new THREE.Vector3(position[0], position[1], position[2]).
		applyMatrix3(matrix.bMatrix);
        vector3 = vector3WorldToView(vector3);
        polygonGroupe.position.set(vector3.x, vector3.y, vector3.z);

        this._mainGroupe.add(polygonGroupe);
    }

    /**
     * This function draw a polygon on each peaks of the scene
     * return => PolygonGroup
     */
    drawMutiple()
    {
        for (let i = peaks.lowX; i <= peaks.upperX; i++) {
            for (let j = peaks.lowY; j <= peaks.upperY; j++) {
                for (let k = peaks.lowZ; k <= peaks.upperZ; k++) {
                    if(!xtal.isPeakAbsent([i,j,k], sample.sg)){
                        this.drawOne([i,j,k]);
                    }
                }
            }
        }       
    }
    

    ////////////////////////////////////////////////////////////
    ////    This part is for all the construction of geometry
    ////////////////////////////////////////////////////////////


    
    /**
     * This function create a geometry from a json file data
     * @returns directly on this._geometry
     */
    createGeometry()
    {
        if(this._jsonData == undefined | this._jsonData == null)
            return;
        let triangles = this._jsonData.polygons;
        let vertices = this._jsonData.vertices;
        this._facesIndex = this._jsonData.faces;

        let newTriangles = [];
        let newVertices = [];

        for (let i = 0; i < triangles.length; i++) 
        {
            for (let j = 0; j < triangles[i].length; j++) 
            {
                newVertices.push(vertices[triangles[i][j]]);
            }

            newTriangles.push([i*3, i*3+1, i*3+2]);
        }

        this._geometry = new THREE.Geometry();

        this.convertArrayToVector3(newVertices, this._geometry.vertices);
        this.convertArrayToFace3(newTriangles, this._geometry.faces);

        this.normalFaceControl(this._geometry);
    }


    /**
     * Dependancy of createGeomtry
     * @param {jsonData array} array 
     * @param {} vertices 
     */
    convertArrayToVector3(array, vertices)
    {
        for (let index = 0; index < array.length; index++) {

            let vertice = new THREE.Vector3(array[index][0], array[index][1], array[index][2]);
            vertice = vector3WorldToView(vertice);
            vertices.push(vertice);
        }
    }

    /**
     * Dependancy of createGeometry
     * @param {jsonData array} array 
     * @param {*} triangles 
     */
    convertArrayToFace3(array, triangles)
    {
        for (let index = 0; index < array.length; index++) {
            triangles.push(new THREE.Face3(array[index][0], array[index][1], array[index][2]));
        }
    }
    /**
     * Dependancy of createGeometry
     * @param {*} geometry 
     */
    normalFaceControl(geometry)
    {
        geometry.computeFaceNormals();
        geometry.computeVertexNormals();

        for (let index = 0; index < geometry.faces.length; index++) {
            let face = geometry.faces[index];
            let vertice = geometry.vertices[face.a];
            
            if(face.normal.x * vertice.x + face.normal.y * vertice.y + face.normal.z * vertice.z < 0.002)
            {
                let newFace = new THREE.Face3(face.a, face.c, face.b);
                newFace.normal = new THREE.Vector3(-face.normal.x, -face.normal.y, -face.normal.z);
                geometry.faces[index] = newFace;
            }
        }
        geometry.computeFaceNormals();
        geometry.computeVertexNormals();

    }

    ////////////////////////////////////////////////////////////
    ////    This part is for get JSON data from takin
    ////////////////////////////////////////////////////////////

    /**
     * This function will get new data from takin_bz
     */

    getNewGeometryData()
    {
        this.createXmlFile();

        let takinTalker = this.takinTalker.bind(this);
        this.checkExistsWithTimeout(".xml", 50, takinTalker);

        let jsonDataReader = this.jsonDataReader.bind(this);
        this.checkExistsWithTimeout(".json", 50, jsonDataReader);
    }


    /**
     * This function take a string of a readed file, and convert it to a json and apply this change to this class
     * @param {string} jsonData 
     */
    jsonDataReader(jsonData)
    {
        this.cleanTemporaryFile("xml");
        this.cleanTemporaryFile("json")

        this._jsonData = JSON.parse(jsonData);
        
        this.createGeometry();
        this.changeBzStruct();
    }

    /**
     * This function create a XML file in vEXP folder
     */
    createXmlFile()
    {
        let operatorsData = xtal.getOperations(sample.sg);
        let operatorString = ""

        for (let ligne = 0; ligne < operatorsData.length; ligne++) {

            let colone = operatorsData[ligne];
            let order = [0, 1, 2, 9, 3, 4, 5, 10, 6, 7, 8, 11];

            operatorString += "<op>"

            for (let index = 0; index < colone.length; index++) {

                operatorString += String(colone[order[index]] + " ")
                
            }
            operatorString += "0 0 0 1 </op>\n";
        }

        let content = '<?xml version="1.0" encoding="utf-8"?>\n\
        <bz>\n\
        <meta>\n\
            <info>bz_tool</info>\n\
        </meta>\n\
        <xtal>\n\
            <a>' + sample.as + '</a>\n\
            <b>' + sample.bs + '</b>\n\
            <c>' + sample.cs + '</c>\n\
            <alpha>' + sample.aa + '</alpha>\n\
            <beta>' + sample.bb + '</beta>\n\
            <gamma>' + sample.cc + '</gamma>\n\
        </xtal>\n\
        <symops>\n\
            ' +operatorString + '\n\
        </symops>\n\
        <formulas/>\n\
        </bz>';

        fs.writeFile(this._fileCodeName + ".xml", content, err => {
            if(err){
                console.log("error");
            }
        });
    }


    /**
     * This function will transfert xml file to takin
     */
    takinTalker()
    {
        exec(this._takinPath + " -c -i " +
		this._fileCodeName + ".xml -o " +
		this._fileCodeName + ".json");
    }

    /**
     * This function will wait untile the file is found, and execute this function with the file as parameter
     * if timeout, function is not call;
     * @param {string} extension 
     * @param {int} timeout 
     * @param {function} callback 
     */
    checkExistsWithTimeout(extension, timeout, callback)
    {
        let name = this._fileCodeName + extension;
        let compteur = 0;
        let intervalObj = setInterval(() => {
            compteur ++;
            if(fs.existsSync(name) && fs.lstatSync(name).isFile())
            {
                clearInterval(intervalObj);
                callback(fs.readFileSync(name, "utf8"));                
                return;
            }
            if(timeout < compteur)
            {
                clearInterval(intervalObj);
                this.cleanTemporaryFile("xml");
                this.cleanTemporaryFile("json")
                return;
            }
        }, 60);
    }

    /**
     * This function create a randome name for temporary file
     * @returns str
     */
    getRandomeFileCodeName()
    {
        let name ="";
        let alphabet = "abcdefghijklmnopqrstuvwxyz"
        for (let index = 0; index < 40; index++) {
            name += String(alphabet[parseInt(Math.random() * 25)]);
        }
        return name;
    }

    /**
     * This Function remove one temporary file
     * @param {str} extention 
     */
    cleanTemporaryFile(extention)
    {
        //fs.unlink(this._fileCodeName + "." + extention);
    }
}

module.exports = BzView3D;
