const THREE = require('three');

const ZOOM_SCALE = 0.90;
const PAN_SPEED = 1;

class MapControls {
    
    constructor(camera, container) {
        this._camera = camera;
        this._container = container;

        this._zoom = 1;
        this._cameraDistance = camera.position.length();

        this._mouse = new THREE.Vector2();
        this._lastMouse = new THREE.Vector2();
        this._enabled = true;

        this._enablePan = false;
        this._enableZoom = false;

        this._listeners = {};

        this._container.addEventListener('mousemove', (event) => {
            this.onMouseMove(event);
        });

        this._container.addEventListener('wheel', (event) => {
            this.onWheel(event);
        });
    }

    set enabled(value) { this._enabled = value; }
    set enablePan(value) { this._enablePan = value; }
    set enableZoom(value) { this._enableZoom = value; }

    get zoom() { return this._zoom; }

    /**
     * Update the control when camera position as changed.
     */
    update() {
        this._zoom = 1;
        this._cameraDistance = this._camera.position.length();
    }

    /**
     * Register a listener on a given event.
     */
    on(event, listener) {
        if(!this._listeners[event]) {
            this._listeners[event] = [listener];
        } else {
            this._listeners[event].push(listener);
        }
    }

    /**
     * Dispatch event and call all listeners.
     */
    dispatch(event, data) {
        if(this._listeners[event]) {
            for(let listener of this._listeners[event]) {
                listener(data);
            }
        }
    }

    /**
     * When the mouse wheel is used, zoom.
     */
    onWheel(event) {

        if (this._enabled && this._enableZoom) {

            let transZ = 0;
            
            // Zoom out
            if (event.deltaY > 0) {
                transZ = this._cameraDistance * (1 - ZOOM_SCALE);
                this._zoom += this._zoom * (1 - ZOOM_SCALE);
            // Zoom in
            } else {
                transZ = -this._cameraDistance * (1 - ZOOM_SCALE);
                this._zoom -= this._zoom * (1 - ZOOM_SCALE);
            }

            this._cameraDistance = this._cameraDistance + transZ;
            this._camera.translateZ(transZ);

            this.dispatch('zoom', { 
                zoom: this._zoom
            });
            this.dispatch('change');
        }
    }

    /**
     * When mouse move and the button is hold, move the camera.
     */
    onMouseMove(event) {
        if(this._enabled && this._enablePan) {
            this._lastMouse.copy(this._mouse);

            let mouseX = event.clientX - $("#" + this._container.id).offset().left;
            let mouseY = event.clientY - $("#" + this._container.id).offset().top;
            this._mouse.set(-mouseX, mouseY);
        
            if(event.buttons != 0) {
                let mouseDelta = this._mouse.clone().sub(this._lastMouse).multiplyScalar(this._zoom * (PAN_SPEED / 4));
                this._camera.translateX(mouseDelta.x);
                this._camera.translateY(mouseDelta.y);
                this.dispatch('change');
            }
        }
    }
}

module.exports = MapControls;
