const { displayMode, receiver } = require('../../../main/context');
const View3D = require('./../view3d');
const THREE = require('three');
const SnapSphere = require('../../../utils/shapes/snap-sphere');
const Arrow3D = require('../../../utils/shapes/arrow-3d');
const Color = require('../../../utils/colors');
const {vector3WorldToView, vector3ViewToWorld } = require('../../../utils/conversion');
const { calculateSs, calculateKiKf } = require('../../../utils/calculations');
const { scattering, matrix, mode } = require('../../../server-state');

class ScatteringView3D extends View3D {
    constructor(scene, dragControls, raycaster, draggables, mapControls) {
        super(scene);

        this._raycaster = raycaster;
        this._origin = new THREE.Vector3(0, 0, 0);
        this._useSnap = true;
        //this._gridVisible = true;

        ////// DRAGGABLE SPHERES //////
        // Draw ki sphere.
        this._kiSphere = new SnapSphere();
        this._kiSphere.mesh.name = "ki";    
        this._kiSphere.position.set(1, 0, 0);
        this._kiSphere.setColor(Color.OK);
        scene.add(this._kiSphere.mesh);
        draggables.push(this._kiSphere.mesh);

        // Draw q sphere.
        this._qSphere = new SnapSphere();
        this._qSphere.mesh.name = "q";
        this._qSphere.position.set(1, 1, 0);
        this._qSphere.setColor(Color.OK);
        scene.add(this._qSphere.mesh);
        draggables.push(this._qSphere.mesh);

        /////// ARROWS //////
        // Rather than moving the arrows, the arrows are tied to the position of two spheres which are draggable.
        // The ki sphere connects the ki arrow and kf arrow.
        // The q sphere connects the q arrow and kf arrow.
        this._kiArrow = new Arrow3D(Color.KI_HIGHLIGHT);
        this._kfArrow = new Arrow3D(Color.KF_HIGHLIGHT);
        this._qArrow = new Arrow3D(Color.Q_HIGHLIGHT);
        
        this._planeArrow = new Arrow3D(0xffffff);

        // Draw ki arrow
        this._kiArrow.object3D.position.copy(this._kiSphere.position);
        this._kiArrow.object3D.lookAt(this._origin);
        this._kiArrow.object3D.scale.z = this._origin.distanceTo(this._kiSphere.position);
        scene.add(this._kiArrow.object3D);

        // Draw kf arrow.
        this._kfArrow.object3D.position.copy(this._kiSphere.position);
        this._kfArrow.object3D.lookAt(this._qSphere.position);
        this._kfArrow.object3D.scale.z = this._qSphere.position.distanceTo(this._kiSphere.position);
        scene.add(this._kfArrow.object3D);

        // Draw q arrow.
        this._qArrow.object3D.position.copy(this._origin);
        this._qArrow.object3D.lookAt(this._qSphere.position);
        this._qArrow.object3D.scale.z = this._qSphere.position.distanceTo(this._origin);
        scene.add(this._qArrow.object3D);

        // Draw plane arrow
        this._planeArrow.object3D.position.copy(this._origin);
        this._planeArrow.object3D.lookAt(new THREE.Vector3(0, 0, 1));
        this._planeArrow.object3D.scale.z = 10;
        scene.add(this._planeArrow.object3D);

        // Drag controls.
        dragControls.addEventListener('dragstart', () => { this.startDrag(); });
        dragControls.addEventListener('drag', (event) => { this.dragArrows(event); });
        dragControls.addEventListener('dragend', () => { this.updatePanel(true); });

        // States for cursor.
        this._dragging = false;
        this._mouseright = false;
        this._ss = 0;

        // Timer to control the nomad access frequency when calling update.
        this._updatePeriodMs = 250;
                        
        // When zooming on 2D view
        mapControls.on('zoom', (event) => {
            let scale = event.zoom * 2;
            this._qSphere.mesh.scale.set(scale, scale, scale);
            this._kiSphere.mesh.scale.set(scale, scale, scale);
        });

        // Activate / Deactivate snap on grid.
        document.addEventListener('mousemove', (event) => {
            this._useSnap = !event.shiftKey;
        });

        document.addEventListener('mousedown', (event) => {
            if (event.button === 2) {
                this._mouseright = true;
            }
        });

        document.addEventListener('mouseup', (event) => {
            if (event.button === 2) {
                this._mouseright = false;
            }
        });

        receiver.on('update-status', (msg, value) => {
            this._setStatus(value.status);
        });
    }

    qPoint() {
        return vector3ViewToWorld(this._qSphere.position);
    }

    kiPoint() {
        return vector3ViewToWorld(this._kiSphere.position);
    }

    /**
     * Draw KI, Kf and Q arrows using Q and Ki spheres.
     */
    drawArrows() {
        this._kiArrow.object3D.position.copy(this._kiSphere.position);
        this._kiArrow.object3D.lookAt(this._origin);
        this._kiArrow.object3D.scale.z = this._kiSphere.position.length();

        this._kfArrow.object3D.position.copy(this._kiSphere.position);
        this._kfArrow.object3D.lookAt(this._qSphere.position);
        this._kfArrow.object3D.scale.z = this._kiSphere.position.distanceTo(this._qSphere.position);

        this._qArrow.object3D.lookAt(this._qSphere.position);
        this._qArrow.object3D.scale.z = this._qSphere.position.length();
    }

    /**
     * Move arrows using the scattering parameters given.
     */
    setScattering(scatteringModel) {

        this._qSphere.position.copy(vector3WorldToView(scatteringModel.kfPoints.end));
        this._kiSphere.position.copy(vector3WorldToView(scatteringModel.kiPoints.start));

        this.drawArrows();
        this._setStatus(scatteringModel.status);

        // For the trajectory view.
        PubSub.publishSync('update-3D-scattering', 0);
    }

    /**
     * Calculate Q value using Q arrow representation.
     */
    getQFromArrow() {
        return vector3ViewToWorld(this._qSphere.position.clone().applyMatrix3(matrix.bMatrix_1));
    }

    /**
     * Project spheres on plane when it moves.
     */
    project() {
        let abVector = new THREE.Vector3(0, 0, 1).applyMatrix3(matrix.uMatrix_1);

        this._kiSphere.position.projectOnPlane(abVector);
        this._qSphere.position.projectOnPlane(abVector);

        this._planeArrow.object3D.lookAt(abVector);

        this.drawArrows();

        // Do not need to update panel in spy mode.
        if (!mode.spyMode) {
            this.updatePanel(true);
        }
    }

    /**
     * Function called when starting the drag.
     */
    startDrag() {

        // Change the state so that the loop continues.
        this._dragging = true;
        this._ss = scattering.ss;
        
        // Start the update drag loop.
        this.updateDragLoop();
    }

    /**
     * Handle event when mouse drag.
     */
    dragArrows(event) {

        let ray = this._raycaster.ray;
        let ab = new THREE.Vector3(0, 0, 1).applyMatrix3(matrix.uMatrix_1);

        // Use the intersectPlane function.
        let newPosition = new THREE.Vector3(0, 0, 0);
        ray.intersectPlane(new THREE.Plane(ab), newPosition);

        let sphere = (event.object.name === "ki") ? this._kiSphere : this._qSphere;
        
        if (this._useSnap && displayMode.value === 'snap') {
            // Spheres movement in snap mode
            sphere.snap(newPosition);
        }
        else {
            // Spheres movement in free mode
            let kiPos = vector3WorldToView(scattering.kiPoints.start.clone());
            let qPos = vector3WorldToView(scattering.kfPoints.end.clone());
            
            if (event.object.name === "q") {
                
                if (scattering.fixedQ) {

                    if (!this._mouseright) {
                        // Ki, Kf and Q are fixed, so we move the triangle.
                        kiPos.applyMatrix3(matrix.uMatrix);
                        qPos.applyMatrix3(matrix.uMatrix);
                        let diffAngle = Math.atan2(kiPos.y, kiPos.x) - Math.atan2(qPos.y, qPos.x);

                        qPos = vector3WorldToView(newPosition.setLength(scattering.q));
                        kiPos = vector3WorldToView(qPos.clone().setLength(scattering.ki)).applyAxisAngle(ab, diffAngle);
                    }
                    else {
                        // trajectory not allowed.
                    }
                }
                else {
                    // We fix Ki and Kf.
                    let q = vector3ViewToWorld(newPosition);
                    let d = scattering.ki + scattering.kf;

                    // If the length of q is greater than ki + kf then we must limit the length of q.
                    if (q.length() > d) {

                        qPos = vector3WorldToView(q.setLength(d));
                        kiPos = vector3WorldToView(qPos.clone().setLength(scattering.ki));
                    }
                    else {
                        // Calculate the new point.
                        let result = calculateKiKf(q, scattering.ki, scattering.kf, this._ss, matrix.uMatrix_1, matrix.uMatrix);
                        
                        qPos = vector3WorldToView(result.kfPoints.end);
                        kiPos = vector3WorldToView(result.kiPoints.start);
                    }
                }

            }
            else if (event.object.name === "ki") {

                if (scattering.fixedDE) {

                    if (!this._mouseright) {
                        // Ki, Kf and Q are fixed, so we move the triangle.
                        kiPos.applyMatrix3(matrix.uMatrix);
                        qPos.applyMatrix3(matrix.uMatrix);
                        let diffAngle = Math.atan2(qPos.y, qPos.x) - Math.atan2(kiPos.y, kiPos.x);

                        kiPos = vector3WorldToView(newPosition.setLength(scattering.ki));
                        qPos = vector3WorldToView(kiPos.clone().setLength(scattering.q)).applyAxisAngle(ab, diffAngle);
                    }
                    else {
                        // trajectory not allowed.
                    }
                }
                else if (scattering.fixedKf) {
                    // Kf is fixed and moving ki sphere.
                    kiPos = vector3WorldToView(newPosition.clone().sub(qPos).setLength(scattering.kf)).add(qPos);
    
                    // Correct the position if we are defining a trajectory (mouse right) and ss is changing.
                    if (this._mouseright) {

                        let kfPoints = {
                            start: vector3ViewToWorld(kiPos),
                            end: vector3ViewToWorld(qPos)
                        };

                        let ss = calculateSs(kfPoints, matrix.uMatrix);
                        
                        if (ss != this._ss) {
                            // Correct the position.
                            let QP = newPosition.clone().sub(qPos);
                            let qn = qPos.clone().normalize();
                            let qd = qn.clone().multiplyScalar(Math.sign(QP.dot(qn)) * scattering.kf);
                            let Q = scattering.kfPoints.end.clone();
                            Q.add(qd);
                            kiPos = vector3WorldToView(Q);
                        }
                    }
                }
                else if (scattering.fixedKi) {
                     // Ki is fixed and moving ki sphere.
                    kiPos = vector3WorldToView(newPosition.clone().setLength(scattering.ki));

                    // Correct the position if we are defining a trajectory (mouse right) and ss is changing.
                    if (this._mouseright) {

                        let kfPoints = {
                            start: vector3ViewToWorld(kiPos),
                            end: vector3ViewToWorld(qPos)
                        };

                        let ss = calculateSs(kfPoints, matrix.uMatrix);
                        
                        if (ss != this._ss) {
                            // Correct the position.
                            let qn = qPos.clone().normalize();
                            let qd = qn.clone().multiplyScalar(Math.sign(kiPos.dot(qn)) * scattering.ki);
                            kiPos = vector3WorldToView(qd);
                        }
                    }
                }
            }
            else {
                // No constraints.
                kiPos = (event.object.name === "ki") ? newPosition : kiPos;
                qPos = (event.object.name === "q") ? newPosition : qPos;
            }

            this._kiSphere.position.copy(kiPos);
            this._qSphere.position.copy(qPos);
        }

        // Update the local data.
        this.updateLocal();

        // Draw the arrows.
        this.drawArrows();
    }

    /**
     * 
     */
    updateLocal() {

        let kiPos = this._kiSphere.position;
        let qPos = this._qSphere.position;
        let qVec = vector3ViewToWorld(qPos).applyMatrix3(matrix.bMatrix_1);

        PubSub.publishSync('update-3D-local', {
            qPoints: qVec, 
            kiPoints: {
                start: vector3ViewToWorld(kiPos),
                end: this._origin.clone()
            },
            kfPoints: {
                start: vector3ViewToWorld(kiPos),
                end: vector3ViewToWorld(qPos)
            },
            dragEnd: true
        });
    }
   
    /**
     * Update Ki, Kf and Q values on panel.
     */
    updatePanel(dragEnd) {
        // If the dragging state is set to false, the drag loop will stop.
        this._dragging = !dragEnd;

        // Calculate and send the points.
        let kiPos = this._kiSphere.position;
        let qPos = this._qSphere.position;
        let qPoints = vector3ViewToWorld(qPos).applyMatrix3(matrix.bMatrix_1);

        let kiPoints = {
            start: vector3ViewToWorld(kiPos),
            end: this._origin.clone()
        }

        let kfPoints = {
            start: vector3ViewToWorld(kiPos),
            end: vector3ViewToWorld(qPos)
        };

        let ss = calculateSs(kfPoints, matrix.uMatrix);

        // Correct ss to the last value if 0.
        if (ss == 0) {
            ss = scattering.ss;
        }

        PubSub.publishSync('update-3D', {
            qPoints,
            kiPoints,
            kfPoints,
            ss
        });
    }

    updateDragLoop() {

        if (this._dragging) {
            this.updatePanel(false);

            setTimeout(this.updateDragLoop.bind(this), this._updatePeriodMs);
        }
    }

    _setStatus(status) {
        if (status === 1) {
            this._qSphere.setColor(Color.OK);
            this._kiSphere.setColor(Color.OK);

            this._qArrow.setColor(Color.Q);
        }
        else {
            this._qSphere.setColor(Color.ERROR);
            this._kiSphere.setColor(Color.ERROR);

            this._qArrow.setColor(Color.ERROR);
        }
    }

}

module.exports = ScatteringView3D;