const viewManager = require('../../utils/view-manager');
const PubSub = require('pubsub-js');
const xtal = require('../../../server/model/xtal/xtal');
const grid = require('ag-grid');
const { atoms } = require('../../server-state');
const atomColor = require('../../utils/atom-color');

function componentToHex(c) {
    var hex = c.toString(16);
    return hex.length == 1 ? "0" + hex : hex;
}
  

class AtomsPanelView {

    constructor() {
    
        let html = `
        <div class='container' style='width:100%'>
            <div class='row'>
                <button class='col-sm-4 btn btn-sm btn-primary' id='addAtom'>Add</button>
                <button class='col-sm-4 btn btn-sm btn-primary' id='delAtom'>Remove</button>
                <button class='col-sm-4 btn btn-sm btn-primary' id='clearAtom'>Clear</button>
            </div>

            <div id='atomgrid' class='ag-theme-fresh' style='height:150px; width:100%; padding-top:10px'/>
        </div>`;


        // Using lambda function does not work here.
        myLayout.registerComponent('Atoms', function(container) {
            container.getElement().html(html);
        });

        viewManager.addView(this);
    }

    init() {
        // Fill in the atom names combo box.
        let atomsList = [ ];
        if(xtal.scatteringLengths !== null) {
            for(let name in xtal.scatteringLengths)
                atomsList.push(name);
        }

        // Create the atoms grid.
        this._atomcols = [
            { headerName: "Name", field: "name", editable: true, cellEditor: "select",
                cellEditorParams: { values: atomsList}, width: 100 },
            { headerName: "x", field: "x", filter: "number", editable: true, width: 100 },
            { headerName: "y", field: "y", filter: "number", editable: true, width: 100 },
            { headerName: "z", field: "z", filter: "number", editable: true, width: 100 },
            { headerName: "Color", field: "color", filter: "number", editable: false, width: 100, cellRenderer: function(params)  {
                let colorString = atomColor(params.data.name, params.data.x).getHexString();
                return `<div style="background-color: #${colorString}; width: 100%">&nbsp</div>`;
            }},
        ];

        this._atomgrid = { columnDefs: this._atomcols, rowData: [],
            enableSorting: true, enableColResize: true, enableFilter: true,
            rowSelection: "single",
            onCellValueChanged: (event) => { this.updateAtoms(event); }
        };
        

        this._theatomgrid = new grid.Grid($("#atomgrid")[0], this._atomgrid);

        let id = 0;
        for (let atom of atoms.atoms) {
            this._atomgrid.api.updateRowData({ add: [{ id, name: atom.name, x: atom.x, y: atom.y, z: atom.z }] });
            id++;
        }

        // Add atom action
        $("#addAtom").click(() => {
            PubSub.publish('add-atom', { name: 'H', x: 0, y: 0, z: 0 });
        });

        // Delete atom action
        $("#delAtom").click(() => {

            // Select the first row.
            let selectedRow = this._atomgrid.api.getSelectedRows()[0];

            // Find the row index to update correctly the model and local atoms data.
            let rowIndex = -1;
            this._atomgrid.api.forEachNode((row, index) => {

                // Test the row by reference.
                if (row.data === selectedRow) {
                    rowIndex = index;
                }
            })

            // The row is found and we pass the selected row by reference (that is removed thereafter).
            if (rowIndex != -1) {
                PubSub.publish('del-atom', { index: rowIndex, row: selectedRow });
            }
        });

        // Clear Atom action
        $("#clearAtom").click(() => {
            PubSub.publish('clear-atoms', 0);
        });
    }

    addAtom(name, x, y, z) {
        let newRow = { name: name, x: x, y: y, z: z };
        this._atomgrid.api.updateRowData({ add: [newRow] });
    }

    delAtom(row) {
        // The reference passed by 'del-atom' is used to correctly identify the row.
        this._atomgrid.api.updateRowData({ remove: [row] });
    }

    clearAtoms() {
        this._atomgrid.api.setRowData([ ]);
    }

    updateAtoms(event) {
        let newAtom = { name: event.data.name, x: parseFloat(event.data.x), y: parseFloat(event.data.y), z: parseFloat(event.data.z) };

        PubSub.publish('update-atom', { index: event.rowIndex, newAtom: newAtom });
    }

    render() {
    }
}

module.exports = AtomsPanelView;
