const THREE = require('three');
const Sample = require('./model/samples/sample');

/*
 * Class implementing directly the calculations without Nomad access.
 */
class DirectDataAccessor {

    constructor() {
      
        this._as = 5;
        this._bs = 5;
        this._cs = 5;
        this._aa = 90;
        this._bb = 90;
        this._cc = 90;
        this._syncSample = true;
        this._ax = 1;
        this._ay = 0;
        this._az = 0;
        this._bx = 0;
        this._by = 1;
        this._bz = 0;
        this._q_h = 1;
        this._q_k = 0;
        this._q_l = 0;
        this._en = 0;
        this._qm = 0;

        this._fx = 2;
        this._ss = 1;

        this._ki = 1;
        this._kf = 1;

        this._qhScan = 0;
        this._qkScan = 0;
        this._qlScan = 0;
        this._enScan = 0;

        this._dqhScan = 0;
        this._dqkScan = 0;
        this._dqlScan = 0;
        this._denScan = 0;

        this._nbPointsScan = 0;
        this._runningScan = false;

        this._kiScan = 0;
        this._kfScan = 0;

        this._fxScan = 0;
        this._ssScan = 1;

        this._calculationStatus = 1;
        this._calculationError = 0;

        // UB matrix.
        this._uBMatrix = new THREE.Matrix3();

        // Angles.
        this._a1 = 45;
        this._a2 = 50;
        this._a3 = 0;
        this._a3p = 0;
        this._a4 = 0;
        this._a5 = 55;
        this._a6 = 60;

        // Resync.
        this.resync();

        // Callbacks
        this._cb_U_changed = function (val) { };
        this._cb_B_changed = function (val) { };
        
        // Initialize U and B.
        this.recalcUB();
    }

    toNumber(value) {
        if (!isFinite(value) || isNaN(value)) {
            return 0.0;
        }
        return value;
    }

    resync() {
       
        this.B = [];
        this.U = [];
        this._UB_obsolete = true;
    }

    resyncSpy(data) {

        // Set the data.
        if (this._syncSample) {
            this._as = data.as;
            this._bs = data.bs;
            this._cs = data.cs;
            this._aa = data.aa;
            this._bb = data.bb;
            this._cc = data.cc;
        }
        this._ax = data.ax;
        this._ay = data.ay;
        this._az = data.az;
        this._bx = data.bx;
        this._by = data.by;
        this._bz = data.bz;
        this._q_h = data.qh;
        this._q_k = data.qk;
        this._q_l = data.ql;
        this._en = data.en;
        this._qm = data.qm;

        this._fx = data.fx;
        this._ss = data.ss;

        this._ki = data.ki;
        this._kf = data.kf;

        if (data.dTypeScan === "center") {

            let n = (data.nbPointsScan - 1) / 2;

            this._qhScan = data.qhScan - n * data.dqhScan;
            this._qkScan = data.qkScan - n * data.dqkScan;
            this._qlScan = data.qlScan - n * data.dqlScan;
            this._enScan = data.enScan - n * data.denScan;
        }
        else if (data.dTypeScan === "min") {

            this._qhScan = data.qhScan;
            this._qkScan = data.qkScan;
            this._qlScan = data.qlScan;
            this._enScan = data.enScan;
        }

        this._dqhScan = data.dqhScan;
        this._dqkScan = data.dqkScan;
        this._dqlScan = data.dqlScan;
        this._denScan = data.denScan;

        this._nbPointsScan = data.nbPointsScan;
        this._runningScan = data.runningScan;

        this._kiScan = data.kiScan;
        this._kfScan = data.kfScan;

        this._fxScan = data.fxScan;
        this._ssScan = data.ssScan;
        this._nameScan = data.nameScan.toString();
        this._startingTime = data.startingTime;

        this.recalcUB();
    }

    recalcUB() {
        let sample = new Sample("", "", "none",
            this._as, this._bs, this._cs,
            this._aa, this._bb, this._cc,
            [this._ax, this._ay, this._az],
            [this._bx, this._by, this._bz]
        );

        sample.calculateReciprocals();
        sample.calculateUBMatrix();

        this._uBMatrix = sample.ub_matrix;
        this.U = sample.u_matrix.elements;
        this.B = sample.b_matrix.elements;

        this._UB_obsolete = false;
    }

    toDegree(angle) {
        return angle / Math.PI * 180.;
    }

    calculateAngles() {

        // Reset status.
        this._calculationStatus = 1;
        this._calculationError = 0;

        // Calculate Q.
        let Qv = new THREE.Vector3(this._q_h, this._q_k, this._q_l).applyMatrix3(this._uBMatrix);
        let Q = Qv.length();
        let Q2 = Qv.lengthSq();
        let ki2 = this._ki * this._ki;
        let kf2 = this._kf * this._kf;

        // From the instrument, to define somewhere (special panel => split Nomad resync, etc.).
        let dM = 3.355;
        let dA = 3.355;

        if (dM * this._ki < Math.PI) {
            console.log('Unreachable K');

            this._calculationStatus = 8;
            this._calculationError = 0;
        }
    
        this._a1 = this.toDegree(this._ss * Math.asin(Math.PI / (this._ki * dM))); // 7.24
        this._a2 = 2 * this._a1; // 7.25

        let ss = 1; // To determine.
        this._a4 = ss * this.toDegree(Math.acos((ki2 + kf2 - Q2) / (2 * this._ki * this._kf))); // 7.26

        let a = 90.;
        if (Qv.x == 0.) {
            if (Qv.y < 0) {
                a = -90.;
            }
            else {
                a = 90.;
            }
        }
        else {
            a = this.toDegree(Math.atan2(Qv.y, Qv.x));
        }

        this._a3 = -a - this.toDegree(Math.acos(-(kf2 - ki2 - Q2) / (2 * Q * this._ki))); // 7.27

        let sa = 1; // To determine.
        this._a5 = sa * this.toDegree(Math.asin(Math.PI / (this._kf * dA))); // 7.29
        this._a6 = 2 * this._a5; // 7.30
    }

    get xbuPath() { return '/users/thales'; }

    set q_h(value) { this._q_h = value; }
    get q_h() { return this._q_h; }
    set q_k(value) { this._q_k = value; }
    get q_k() { return this._q_k; }
    set q_l(value) { this._q_l = value; }
    get q_l() { return this._q_l; }
    set fx(value) { this._fx = value; }
    get fx() { return this._fx; }

    get bMatrix() {
        if (this._UB_obsolete)
            this.recalcUB();
        return this.B;
    }
    
    get uMatrix() {
        if (this._UB_obsolete)
            this.recalcUB();
        return this.U;
    }

    set as(value) { this._as = value; this._UB_obsolete = true; }
    get as() { return this._as; }
    set bs(value) { this._bs = value; this._UB_obsolete = true; }
    get bs() { return this._bs; }
    set cs(value) { this._cs = value; this._UB_obsolete = true; }
    get cs() { return this._cs; }
    set aa(value) { this._aa = value; this._UB_obsolete = true; }
    get aa() { return this._aa; }
    set bb(value) { this._bb = value; this._UB_obsolete = true; }
    get bb() { return this._bb; }
    set cc(value) { this._cc = value; this._UB_obsolete = true; }
    get cc() { return this._cc; }

    set syncSample(value) {
        this._syncSample = value;
    }

    set ax(value) { this._ax = value; this._UB_obsolete = true; }
    get ax() { return this._ax; }
    set ay(value) { this._ay = value; this._UB_obsolete = true; }
    get ay() { return this._ay; }
    set az(value) { this._az = value; this._UB_obsolete = true; }
    get az() { return this._az; }
    set bx(value) { this._bx = value; this._UB_obsolete = true; }
    get bx() { return this._bx; }
    set by(value) { this._by = value; this._UB_obsolete = true; }
    get by() { return this._by; }
    set bz(value) { this._bz = value; this._UB_obsolete = true; }
    get bz() { return this._bz; }

    set ki(value) { this._ki = value; }
    get ki() { return this._ki; }

    set kf(value) { this._kf = value; }
    get kf() { return this._kf; }

    get ss() { return this._ss; }

    set calcValues(value) {
        let values = JSON.parse(value);

        this._q_h = values.qh;
        this._q_k = values.qk;
        this._q_l = values.ql;
        this._ki = values.ki;
        this._kf = values.kf;
        this._fx = values.fx;
        this._ss = values.ss;

        this.calculateAngles();
    }

    get a1() { return this._a1; }
    get a2() { return this._a2; }
    get a3() { return this._a3; }
    get a3p() { return this._a3p; }
    get a4() { return this._a4; }
    get a5() { return this._a5; }
    get a6() { return this._a6; }

    get calculationStatus() {
        return this._calculationStatus;
    }

    get calculationError() {
        return this._calculationError;
    }

    get qhScan() {
        return this._qhScan;
    }

    get qkScan() {
        return this._qkScan;
    }

    get qlScan() {
        return this._qlScan;
    }

    get enScan() {
        return this._enScan;
    }

    get dqhScan() {
        return this._dqhScan;
    }

    get dqkScan() {
        return this._dqkScan;
    }

    get dqlScan() {
        return this._dqlScan;
    }

    get denScan() {
        return this._denScan;
    }

    get kiScan() {
        return this._kiScan;
    }

    get kfScan() {
        return this._kfScan;
    }

    get fxScan() {
        return this._fxScan;
    }

    get ssScan() {
        return this._ssScan;
    }

    get nbPointsScan() {
        return this._nbPointsScan;
    }

    get runningScan() {
        return this._runningScan;
    }

    get nameScan() {
        return this._nameScan;
    }

    get startingTime() {
        return this._startingTime;
    }
}

// We export the constructor function.
module.exports = DirectDataAccessor;