/**
 * cif handling
 * @author Tobias Weber <tweber@ill.fr>
 * @date Mar-2019
 * @license see "LICENSE" file
 */

//"use strict";

const eps = 1e-5;
const mathjs = require("mathjs");
const math = mathjs.create(mathjs.all, {epsilon: eps});


class Cif 
{
    constructor(options = {}) {
        this._cifhandle = -1;

        this._nativelib = require("../../vexplib");
        if(this._nativelib === null) {
            return;
        }

        this._cifhandle = this._nativelib.createCIF();
    }


    load(filedata) {
        if(this._nativelib === null || this._cifhandle < 0) {
            console.error("Invalid CIF handle, vexplib is not loaded (correctly).");
            return false;
        }

        this._nativelib.clearCIF(this._cifhandle);
        return this._nativelib.loadCIF(this._cifhandle, filedata, false);
    }


    getLattice() {
        if(this._nativelib === null || this._cifhandle < 0) {
            console.error("Invalid CIF handle, vexplib is not loaded (correctly).");
            return false;
        }

        return this._nativelib.getCIFLattice(this._cifhandle);
    }


    getAtoms() {
        if(this._nativelib === null || this._cifhandle < 0) {
            console.error("Invalid CIF handle, vexplib is not loaded (correctly).");
            return [];
        }

        const numatoms = this._nativelib.getCIFNumAtoms(this._cifhandle);
        //console.log("Numer of atoms: " + numatoms);

        let allatoms = [];
        for(let atomidx=0; atomidx<numatoms; ++atomidx) {
            const theatom = this._nativelib.getCIFAtom(this._cifhandle, atomidx);
            allatoms.push(theatom);
        }

        return allatoms;
    }


    getGeneratedAtoms() {
        if(this._nativelib === null || this._cifhandle < 0) {
            console.error("Invalid CIF handle, vexplib is not loaded (correctly).");
            return [];
        }

        const numatoms = this._nativelib.getCIFNumAtoms(this._cifhandle);

        let allatoms = [];
        for(let atomidx=0; atomidx<numatoms; ++atomidx) {
            const numgenatoms = this._nativelib.getCIFNumGeneratedAtoms(this._cifhandle, atomidx);

            for(let genatomidx=0; genatomidx<numgenatoms; ++genatomidx) {
                const theatom = this._nativelib.getCIFGeneratedAtom(this._cifhandle, atomidx, genatomidx);
                allatoms.push(theatom);
            }
        }

        return allatoms;
    }


    getSymOps() {
        if(this._nativelib === null || this._cifhandle < 0) {
            console.error("Invalid CIF handle, vexplib is not loaded (correctly).");
            return [];
        }

        const numops = this._nativelib.getCIFNumSymOps(this._cifhandle);
        //console.log("Numer of symops: " + numops);

        let allops = [];
        for(let opidx=0; opidx<numops; ++opidx) {
            const theop = this._nativelib.getCIFSymOp(this._cifhandle, opidx);
            const symop = theop.op;
            //console.log("Symop: " + symop);

            const M = math.matrix([
                [ symop[0], symop[1], symop[2], symop[3] ],
                [ symop[4], symop[5], symop[6], symop[7] ],
                [ symop[8], symop[9], symop[10], symop[11] ],
                [ symop[12], symop[13], symop[14], symop[15] ]]);
            allops.push(M);
        }

        return allops;
    }
};


// ----------------------------------------------------------------------------
/**
 * exported data and functions
 */
module.exports =
{
    "Cif" : Cif,
};
// ----------------------------------------------------------------------------
