/*
 * Class delegating the calculations to Nomad or not.
 * It delegates either to a NomadDataAccessor object or DirectDataAccessor object.
 */

const PubSub = require('pubsub-js');
const THREE = require('three');
const NomadDataAccessor = require('./nomad-data-accessor');
const DirectDataAccessor = require('./direct-data-accessor');
const vEXPSpy = require('./vexp-spy');
const { sender } = require('./context');
const modeModel = require('./model/mode-model');

class VEXPController {

    constructor() {

        // Data accessor.
        if (modeModel.nomadConnection) {
            this._dataAccessor = new NomadDataAccessor(modeModel.nomadId);
            console.log('Connection to Nomad');
        }
        else {
            this._dataAccessor = new DirectDataAccessor();
            console.log('No connection to Nomad');
        }

        this._as = 5;
        this._bs = 5;
        this._cs = 5;
        this._aa = 90;
        this._bb = 90;
        this._cc = 90;
        this._ax = 1;
        this._ay = 0;
        this._az = 0;
        this._bx = 0;
        this._by = 1;
        this._bz = 0;

        this._uMatrix = new THREE.Matrix3();
        this._uMatrix_1 = new THREE.Matrix3();
        this._bMatrix = new THREE.Matrix3();
        this._bMatrix_1 = new THREE.Matrix3();
        this._uBMatrix = new THREE.Matrix3();
        this._uBMatrix_1 = new THREE.Matrix3();

        this._ki = this._dataAccessor.ki;
        this._kf = this._dataAccessor.kf;
        this._q_h = this._dataAccessor.q_h;
        this._q_k = this._dataAccessor.q_k;
        this._q_l = this._dataAccessor.q_l;
        this._fx = this._dataAccessor.fx;
        this._ss = this._dataAccessor.ss;
        this._sg = "unknown";

        this._calculationStatus = 0;
        this._calculationError = 0;

        this._xbuPath = this._dataAccessor.xbuPath;

        this._a1 = 0;
        this._a2 = 0;
        this._a3 = 0;
        this._a3p = 0;
        this._a4 = 0;
        this._a5 = 0;
        this._a6 = 0;

        this._qhScan = 0;
        this._qkScan = 0;
        this._qlScan = 0;
        this._enScan = 0;
        this._dqhScan = 0;
        this._dqkScan = 0;
        this._dqlScan = 0;
        this._denScan = 0;

        this._kiScan = 0;
        this._kfScan = 0;
        this._fxScan = 1;
        this._ssScan = 1;

        this._nbPointsScan = 0;
        this._runningScan = false;
        this._nameScan = "";
        this._startingTime = "";
        this._dataScan = [];

        // Spy.
        if (modeModel.spyMode) {
            vEXPSpy.init();
            this._loopSpyUpdate();
        }
    }

    /**
     * Resynchronize vEXP with Nomad.
     */
    resync() {
        this._dataAccessor.resync();
        this._ki = this._dataAccessor.ki;
        this._kf = this._dataAccessor.kf;
        this._q_h = this._dataAccessor.q_h;
        this._q_k = this._dataAccessor.q_k;
        this._q_l = this._dataAccessor.q_l;
        this._fx = this._dataAccessor.fx;
        this._ss = this._dataAccessor.ss;

        this._xbuPath = this._dataAccessor.xbuPath;
        this.initValues();
        this.updateMatrix();
    }

    resyncSpy() {
        let data = vEXPSpy.getData();

        if (data == null) {
            return;
        }

        this._dataAccessor.resyncSpy(data);

        // Check if the trajectories have to be cleared.
        //let clearTrajectories = (this._as !== this._dataAccessor.as || this._bs !== this._dataAccessor.bs || this._cs !== this._dataAccessor.cs || this._aa !== this._dataAccessor.aa || this._bb !== this._dataAccessor.bb || this._cc !== this._dataAccessor.cc);

        // Set the data.
        this._as = this._dataAccessor.as;
        this._bs = this._dataAccessor.bs;
        this._cs = this._dataAccessor.cs;
        this._aa = this._dataAccessor.aa;
        this._bb = this._dataAccessor.bb;
        this._cc = this._dataAccessor.cc;
        this._ax = this._dataAccessor.ax;
        this._ay = this._dataAccessor.ay;
        this._az = this._dataAccessor.az;
        this._bx = this._dataAccessor.bx;
        this._by = this._dataAccessor.by;
        this._bz = this._dataAccessor.bz;
        this._q_h = this._dataAccessor.q_h;
        this._q_k = this._dataAccessor.q_k;
        this._q_l = this._dataAccessor.q_l;
        this._en = this._dataAccessor.en;
        
        this._fx = this._dataAccessor.fx;
        this._ss = this._dataAccessor.ss;

        this._ki = this._dataAccessor.ki;
        this._kf = this._dataAccessor.kf;

        this._qhScan = this._dataAccessor.qhScan;
        this._qkScan = this._dataAccessor.qkScan;
        this._qlScan = this._dataAccessor.qlScan;
        this._enScan = this._dataAccessor.enScan;

        this._dqhScan = this._dataAccessor.dqhScan;
        this._dqkScan = this._dataAccessor.dqkScan;
        this._dqlScan = this._dataAccessor.dqlScan;
        this._denScan = this._dataAccessor.denScan;

        this._nbPointsScan = this._dataAccessor.nbPointsScan;
        this._runningScan = this._dataAccessor.runningScan;

        this._kiScan = this._dataAccessor.kiScan;
        this._kfScan = this._dataAccessor.kfScan;

        this._fxScan = this._dataAccessor.fxScan;
        this._ssScan = this._dataAccessor.ssScan;

        if (this._dataAccessor.nameScan === "0") {
            this._nameScan = "?";
        }
        else {
            this._nameScan = this._dataAccessor.nameScan;
        }

        this._startingTime = this._dataAccessor.startingTime;
        this._dataScan = data.dataScan;

        // Update the matrix.
        this.updateMatrix();

        // Notify the main.
        PubSub.publishSync('syncSpy', false);
    }

    _loopSpyUpdate() {
        this.resyncSpy();
        setTimeout(this._loopSpyUpdate.bind(this), 1000);
    }

    initValues() {

        // Initialize the sample parameters.
        /*this._as = this._dataAccessor.as;
        this._bs = this._dataAccessor.bs;
        this._cs = this._dataAccessor.cs;
        this._aa = this._dataAccessor.aa;
        this._bb = this._dataAccessor.bb;
        this._cc = this._dataAccessor.cc;*/

        // Initialize the plane parameters.
        this._ax = this._dataAccessor.ax;
        this._ay = this._dataAccessor.ay;
        this._az = this._dataAccessor.az;
        this._bx = this._dataAccessor.bx;
        this._by = this._dataAccessor.by;
        this._bz = this._dataAccessor.bz;

        this.updateMatrix();

        this._dataAccessor.ki = this._ki;
        this._dataAccessor.kf = this._kf;
        this._dataAccessor.q_h = this._q_h;
        this._dataAccessor.q_k = this._q_k;
        this._dataAccessor.q_l = this._q_l;
        this._dataAccessor.fx = this._fx;

        // Update the calculation, status and angles.
        this.update();

        // space groups and atoms
        this._atomnames = null;
        this._atompos = null;
    }

    updateMatrix() {
        this._uMatrix.fromArray(this._dataAccessor.uMatrix);
        this._uMatrix_1.getInverse(this._uMatrix);

        this._bMatrix.fromArray(this._dataAccessor.bMatrix);
        this._bMatrix_1.getInverse(this._bMatrix);

        this._uBMatrix.multiplyMatrices(this._uMatrix, this._bMatrix);
        this._uBMatrix_1.getInverse(this._uBMatrix);
    }

    // Send the plane values to Nomad to get new U and M matrices.
    setPlanes(ax, ay, az, bx, by, bz) {
        this._ax = ax;
        this._ay = ay;
        this._az = az;
        this._bx = bx;
        this._by = by;
        this._bz = bz;

        this._dataAccessor.ax = this._ax;
        this._dataAccessor.ay = this._ay;
        this._dataAccessor.az = this._az;
        this._dataAccessor.bx = this._bx;
        this._dataAccessor.by = this._by;
        this._dataAccessor.bz = this._bz;

        this.updateMatrix();
    }

    setQFromUI(h, k, l) {
        this._q_h = h;
        this._q_k = k;
    }

    setSampleFromUI(as, bs, cs, aa, bb, cc, sg) {
        this._dataAccessor.as = this._as = as;
        this._dataAccessor.bs = this._bs = bs;
        this._dataAccessor.cs = this._cs = cs;
        this._dataAccessor.aa = this._aa = aa;
        this._dataAccessor.bb = this._bb = bb;
        this._dataAccessor.cc = this._cc = cc;
        this._sg = sg;

        this.updateMatrix();
    }

    setAtomsFromUI(atomnames, atompos) {
        this._atomnames = atomnames;
        this._atompos = atompos;
    }

    get syncSample() {
        return this._dataAccessor._syncSample;
    }

    set syncSample(value) {
        this._dataAccessor._syncSample = value;

        // If sync is true, We must reset the space group.
        if (value) {
            this._sg = "unknown";
        }
    }

    /**
     * Send calculation command to nomad and retrieve calculated
     * angles.
     */
    update() {

        // Do not calculate in spy mode.
        if (modeModel.spyMode) {
            return;
        }

        let calcValues = {
            qh: this._q_h,
            qk: this._q_k,
            ql: this._q_l,
            ki: this._ki,
            kf: this._kf,
            fx: this._fx,
            ss: this._ss
        };

        // Set the value triggers the calculation in Nomad.
        this._dataAccessor.calcValues = JSON.stringify(calcValues);

        // Update the status and error.
        this._calculationStatus = parseInt(this._dataAccessor.calculationStatus);
        this._calculationError = parseInt(this._dataAccessor.calculationError);

        // Get the angles.
        this._a1 = this._dataAccessor.a1;
        this._a2 = this._dataAccessor.a2;
        this._a3 = this._dataAccessor.a3;
        this._a3p = this._dataAccessor.a3p;
        this._a4 = this._dataAccessor.a4;
        this._a5 = this._dataAccessor.a5;
        this._a6 = this._dataAccessor.a6;

        // Update model, notify it that angles can be retrieved.
        PubSub.publish('update-angles', 0);
        
        // Inform client when not in spy mode.
        sender.send('update-angles', {
            a1: this._a1,
            a2: this._a2,
            a3: this._a3,
            a3p: this._a3p,
            a4: this._a4,
            a5: this._a5,
            a6: this._a6
        });
        
        sender.send('update-status', {
            status: this._calculationStatus,
            error: this._calculationError
        });
    }

    // -------------------------------------------------------------------------
    // getters & setters

    get xbuPath() { return this._xbuPath; }

    get a1() { return this._a1; }
    get a2() { return this._a2; }
    get a3() { return this._a3; }
    get a3p() { return this._a3p; }
    get a4() { return this._a4; }
    get a5() { return this._a5; }
    get a6() { return this._a6; }

    get ki() { return this._ki; }
    set ki(value) { this._ki = value; }

    get kf() { return this._kf; }
    set kf(value) { this._kf = value; }

    get q_h() { return this._q_h; }
    set q_h(value) { this._q_h = value; this._dataAccessor.q_h = value; }

    get q_k() { return this._q_k; }
    set q_k(value) { this._q_k = value; this._dataAccessor.q_k = value; }

    get q_l() { return this._q_l; }
    set q_l(value) { this._q_l = value; this._dataAccessor.q_l = value; }

    get fx() { return this._fx; }
    set fx(value) { this._fx = value; this._dataAccessor.fx = value; }
    
    get ss() { return this._ss; }
    set ss(value) { this._ss = value; }

    get bMatrix() { return this._bMatrix; }
    set bMatrix(value) { this._bMatrix = value; }

    get uMatrix() { return this._uMatrix; }
    set uMatrix(value) { this._uMatrix = value; }

    get uBMatrix() { return this._uBMatrix; }
    set uBMatrix(value) { this._uBMatrix = value; }

    get as() { return this._as; }
    set as(value) { this._as = value; }

    get bs() { return this._bs; }
    set bs(value) { this._bs = value; }

    get cs() { return this._cs; }
    set cs(value) { this._cs = value; }

    get aa() { return this._aa; }
    set aa(value) { this._aa = value; }

    get bb() { return this._bb; }
    set bb(value) { this._bb = value; }

    get cc() { return this._cc; }
    set cc(value) { this._cc = value; }

    get ax() { return this._ax; }
    set ax(value) { this._ax = value; }

    get ay() { return this._ay; }
    set ay(value) { this._ay = value; }

    get az() { return this._az; }
    set az(value) { this._az = value; }

    get bx() { return this._bx; }
    set bx(value) { this._bx = value; }

    get by() { return this._by; }
    set by(value) { this._by = value; }

    get bz() { return this._bz; }
    set bz(value) { this._bz = value; }

    get sg() { return this._sg; }
    set sg(value) { this._sg = value; }

    get atomnames() { return this._atomnames; }
    set atomnames(value) { this._atomnames = value; }

    get atompos() { return this._atompos; }
    set atompos(value) { this._atompos = value; }

    get uBMatrix_1() { return this._uBMatrix_1; }
    set uBMatrix_1(value) { this._uBMatrix_1 = value; }

    get uMatrix_1() { return this._uMatrix_1; }
    set uMatrix_1(value) { this._uMatrix_1 = value; }

    get bMatrix_1() { return this._bMatrix_1; }
    set bMatrix_1(value) { this._bMatrix_1 = value; }

    get calculationStatus() { return this._calculationStatus; }

    get qhScan() { return this._qhScan; }
    get qkScan() { return this._qkScan; }
    get qlScan() { return this._qlScan; }
    get enScan() { return this._enScan; }
    get dqhScan() { return this._dqhScan; }
    get dqkScan() { return this._dqkScan; }
    get dqlScan() { return this._dqlScan; }
    get denScan() { return this._denScan; }
    get kiScan() { return this._kiScan; }
    get kfScan() { return this._kfScan; }
    get fxScan() { return this._fxScan; }
    get ssScan() { return this._ssScan; }
    get nbPointsScan() { return this._nbPointsScan; }
    get runningScan() { return this._runningScan; }
    get nameScan() { return this._nameScan; }
    get startingTime() { return this._startingTime; }
    get dataScan() { return this._dataScan; }
}

const vEXPController = new VEXPController();

module.exports = vEXPController;
